/* Copyright (C) 2003 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
 
#include <includes.h>
#define TYPE_TO_MASTER		1


/*-----------------------[ CLIENT PART ]------------------------------------*/

	
	

static struct http_msg * recv_http_msg(int msgid, int type)
{
 struct http_msg * msg;
 int allocated_size = MSGMAX;
 int n; 
 
 
 msg = emalloc(allocated_size);
 
 
again:
 n = msgrcv (msgid, msg, allocated_size, type, 0 );
 if ( n < 0 && errno == EINTR )
  goto again;
 else if( n < 0 )
 {
  efree(&msg);
  return NULL;
 }
 if(msg->total_len != msg->data_len)
 {
  struct http_msg * unfragmented = emalloc( sizeof(struct http_msg) + msg->total_len ) ;
  memcpy(unfragmented, msg, sizeof(struct http_msg) + msg->data_len);
  
  while(unfragmented->data_len != msg->total_len )
  {
again2:  
    n = msgrcv (msgid, msg, allocated_size,type, 0 );
    if(n < 0 && errno == EINTR) goto again2;
    
    memcpy(unfragmented->data + unfragmented->data_len, msg->data, msg->data_len);
    unfragmented->data_len += msg->data_len;
  }
  efree(&msg);
  return unfragmented;
 }
 else
  return msg;
}

static int send_http_msg(int msgid, struct http_msg * msg)
{
 int n;
 
 msg->total_len = msg->data_len;
 
 
 if( msg->total_len > 0 && msg->total_len >= MSGMAX )
 {
  struct http_msg * helper = emalloc(MSGMAX);
  int sent = 0, total = msg->total_len;

  memcpy(helper, msg, sizeof(struct http_msg));
  while (sent != total )
  {
   int m;
   
   if( total - sent > (MSGMAX - sizeof(struct http_msg)))
    m = (MSGMAX) - sizeof(struct http_msg);
   else
    m = total - sent;
    
   memcpy(helper->data, msg->data + sent, m);
   helper->data_len = m;
again2:
   n =  msgsnd(msgid, helper, helper->data_len + sizeof(struct http_msg), 0);
   if( n < 0 && errno == EINTR )
   	goto again2;
   else if( n < 0 )
   {
    perror("msgsnd ");
    efree(&helper);
    return -1;
   }	
	
   sent += m;
  }
  
  efree(&helper);
  return 0;
 }
 else
 {
sndagain :
 n = msgsnd(msgid, msg, msg->data_len >0 ? sizeof(struct http_msg) + msg->data_len:sizeof(struct http_msg), 0 );
 if( n < 0 && errno == EINTR) goto sndagain;
 else if ( n < 0 )return -1; 
 }
 return 0;
}


struct http_msg * http_share_mkmsg(int port, int transport, int data_len,char * data)
{
 struct http_msg * ret = emalloc(sizeof(struct http_msg) + data_len);
 
 ret->type = TYPE_TO_MASTER;
 ret->owner = getpid();
 ret->port = port;
 ret->transport = transport;
 ret->data_len = data_len;
 ret->total_len = data_len;
 memcpy(ret->data, data, data_len);

 return ret;
}


void http_share_freemsg(struct http_msg * msg)
{
 efree(&msg);
}




int http_share_exists(struct arglist * args)
{
 int msgid;
 struct in_addr * host;
 
 host = plug_get_host_ip(args);
 if ( host == NULL ) return -1;
 
 
 msgid = msgget(host->s_addr, 0);
 if( msgid < 0 )
 	return -1;
 else
 	return 0;
}



struct http_msg * http_share_send_recv_msg(struct arglist * args, struct http_msg * sndmsg)
{
 int msgid;
 int n;
 struct in_addr * host;
 struct http_msg * msg;
 
 host = plug_get_host_ip(args);
 if ( host == NULL ) return NULL;
 
 
 msgid = msgget(host->s_addr, 0);
 if( msgid < 0 )
 { 
  /* Ack! What's going on ?! */
  fprintf(stderr, "libnessus:http_share_send_recv_msg(): Could not connect to handler ?!\n");
  return NULL;
 }
 
 send_http_msg(msgid, sndmsg);
 msg = recv_http_msg(msgid, getpid());
 return msg;
}

/*--------------------------[ SERVER SIDE ]----------------------------------*/

struct established_socket {
	int port;
	int fd;
	int transport;
	};

#define MAX_SOCKS 32
static struct established_socket sockets[MAX_SOCKS];
static struct arglist * host_args = NULL;




static int get_sock(int port)
{
 int i;
 for( i = 0; i < MAX_SOCKS ; i ++ )
 {
  if( sockets[i].port == port )
  	return sockets[i].fd;

  else if( sockets[i].port == 0 )
  	break;	
 } 
 
 return -1;
}


static int new_sock(int port, int transport)
{
 int i;
 for( i = 0 ;i< MAX_SOCKS ; i ++ )
 {
  if ( sockets[i].port == 0 )
  	break;
 }
 
 if( i != MAX_SOCKS )
 {
  sockets[i].port = port;
  sockets[i].transport = transport;
  sockets[i].fd = open_stream_connection(host_args, port, transport, 5);
  if( sockets[i].fd <  0)
  {
   sockets[i].port = 0;
   return -1;
  }
  return sockets[i].fd;
 }
 else return -1;
}


static void rm_sock(int fd)
{
 int i;
 for(i=0;i<MAX_SOCKS;i++)
 {
  if(sockets[i].fd == fd )
  {
   close_stream_connection(fd);
   bzero(&(sockets[i]), sizeof(sockets[i]));
   return;
  }
  else if(sockets[i].port == 0)
  	break; /* ?? */
 }
}


static int reopen_sock(int fd)
{
 int i;
 for(i=0;i<MAX_SOCKS;i++)
 {
  if(sockets[i].fd == fd )
  {
   close_stream_connection(fd);
   sockets[i].fd = open_stream_connection(host_args, sockets[i].port, sockets[i].transport, 5);
   if(sockets[i].fd < 0 )
   { 
    bzero(&(sockets[i]), sizeof(sockets[i]));
    return -1;
   }
   else return sockets[i].fd;
  }
  else if(sockets[i].port == 0)
  	break; /* ?? */
 }
}


static int extract_content_length(char * headers)
{
 char * str;
 str = strstr(headers, "\nContent-Length: ");
 if( str != NULL )
 {
  str += strlen("\nContent-Length: ");
  return atoi(str);
 }
 
 return -1;
}

static int chunked_encoding(char * headers)
{
 if(strstr(headers, "Transfer-Encoding: chunked") != NULL )
  return 1;
 else 
  return 0;
}


static int recv_chunked_encoding(int fd, char * headers, int hlen, char ** buf, int * len)
{
 int total_len, sz = hlen;
 char tmp[2048];
 char * mybuf = emalloc(hlen);
 

 memcpy(mybuf, headers, hlen);
 for(;;)
 { 
  int n;
  char * end;
  tmp[sizeof(tmp) - 1] = '\0';
  n = recv_line(fd, tmp, sizeof(tmp) - 1);
  if (n <= 0) break;
  total_len = strtol(tmp, &end, 16);
  
  mybuf = erealloc(mybuf, sz + total_len + 2);
  n = read_stream_connection_min(fd, mybuf + sz, total_len + 2, total_len + 2);
  if( n <= 0 || total_len == 0 || sz > 5 * 1024 * 1024)
   {
   if( sz > 5 * 1024 * 1024 )fd = reopen_sock(fd);
   break;
   }
    sz += total_len + 2;
  }
  
   *len = sz;
   *buf = mybuf;
   return 0;
}


static char * http_share_recv(int fd, int * totlen, int was_head)
{
 char * headers;
 int headers_len;
 int close_me = 0;
 
 *totlen = 0;
 http_recv_headers(fd, &headers, &headers_len);
	
 if(headers_len <= 0 || headers == NULL )
 	return NULL;
	
  if(strstr(headers, "\nConnection: close") != NULL ||
     strstr(headers, "\nConnection: Close") != NULL)
     	{
    	close_me = 1;
	}
		
		
 if( was_head != 0 )
 { 
  *totlen = headers_len;
  if(close_me)reopen_sock(fd);
  return headers;
 }		
	

 if( headers != NULL  )
 {
  int len = extract_content_length(headers);
  
  if ( len >= 0  )
  {
   char * retbuf;
   
   /* Do not allow a too long response. */
   if( len > 1024*1024 )
   {
    len = 1024*1024;
    close_me = 1;
   }
   
   retbuf = emalloc(headers_len + len);
   memcpy(retbuf, headers, headers_len);
   read_stream_connection_min(fd, retbuf + headers_len, len, len);
   *totlen = len + headers_len;
   if( close_me != 0 ) reopen_sock(fd);
   efree(&headers);
   return retbuf;
  }
  else if (chunked_encoding(headers))
  {
   char * ret_buf;
   int len;

   recv_chunked_encoding(fd, headers, headers_len, &ret_buf, &len);
   *totlen = len;
   if( close_me != 0 ) reopen_sock(fd);
   efree(&headers);
   return ret_buf;
  }
  else
  {
   int len = headers_len + 16384;
   char * buf = emalloc(len);
   int n;
   
   memcpy(buf, headers, headers_len);
   n = read_stream_connection(fd, buf + headers_len, 16384);
   if( n >= 0 )n += headers_len;
   else n = headers_len;
   reopen_sock(fd); /* Alway close */
   *totlen = n;
   efree(&headers);
   return buf;
  }
 }
 else return NULL;
}

static int http_share_process(int msgid, struct http_msg * message, struct http_msg ** reply, int * allocated_memory)
{
 int soc = get_sock(message->port);
 int n;
 int flag = 0;
 int was_head;
 

 
 
 if( soc < 0 )
 { 
  soc = new_sock(message->port, message->transport);
  if( soc < 0 ) 
   { 
    *reply = emalloc(sizeof(struct http_msg));
    memcpy(*reply, message, sizeof(struct http_msg));
    (*reply)->data_len = -1;
    
    return 0;
   }
 }
 
 /* 
  * If this is a HEAD request, force the shutdown of the connection, as we will
  * obtain a bogus length field
  */
 if(strncmp(message->data, "HEAD", 4) == 0 ) was_head = 1;
 else was_head = 0;
 
 
again: 
 /* Send the HTTP request */
 n = write_stream_connection(soc, message->data, message->data_len);
 if ( n <= 0 )
 { 
  if( flag == 0 )
  {
   soc = reopen_sock(soc);
   if( soc >= 0 )
   {
    flag = 1;
    goto again;
   }
  }
  
  (*reply)->data_len = -1;
 }
 else
 {
  /* Receive the message */
  int len;
  char * buf = http_share_recv(soc, &len, was_head);
  
  if ( buf == NULL ) 
  	len = 0;
  
  *reply = emalloc(len + sizeof(struct http_msg));
  memcpy(*reply, message, sizeof(struct http_msg));
  *allocated_memory = len + sizeof(struct http_msg);
  
    
   if( buf != NULL )
   	(*reply)->data_len = len;
   else
   	(*reply)->data_len = -1;
	
   if ( buf != NULL && len > 0) memcpy((*reply)->data, buf, len);
   efree(&buf);
 }
 return 0;
}


/*----------------------- Public server functions --------------------*/

int http_share_close(struct arglist * args, pid_t pid)
{
 int msgid;
 struct in_addr * addr;
 
 
 kill(pid, SIGTERM);
 
 
 addr = plug_get_host_ip(args);
 if( addr == NULL ) return -1;
 
 
 msgid = msgget(addr->s_addr, 0);
 if( msgid < 0 ) return -1;
 msgctl(msgid, IPC_RMID, NULL);
 return 0;
}


pid_t http_share_init(struct arglist * hostdata)
{
 pid_t pid;
 int msgid;
 struct http_msg * msg;
 struct http_msg * reply;
 int reply_sz;
 struct in_addr * addr;
 char title[512];
 
 
 pid = fork();
 if( pid != 0 )return pid;
 signal(SIGTERM, _exit);

 host_args = hostdata;
 
 addr = plug_get_host_ip(hostdata);
 snprintf(title, sizeof(title), "http sharer for %s", inet_ntoa(*addr));
 setproctitle("%s", title);
 bzero(sockets, sizeof(sockets));
 msgid = msgget(addr->s_addr, 0);
 if( msgid >= 0 )
 {
  msgctl(msgid, IPC_RMID, NULL);
 }
 
 
 
 msgid = msgget(addr->s_addr, IPC_CREAT|0600);
 if( msgid < 0 )
 {
  perror("msgget() ");
  exit (1);
 }
 
 
 
 for(;;)
 {
  int n;
  msg = recv_http_msg(msgid, TYPE_TO_MASTER);
  if( msg == NULL )
  {
   perror("msgrcv() ");
   exit(1);
  }

  /* Process the message */
  if(http_share_process(msgid, msg, &reply, &reply_sz) == 0 )
  {
   int n;
   
 
   reply->type = msg->owner;
   
   send_http_msg(msgid, reply);
   http_share_freemsg(reply);
  }
  efree(&msg);  
 }
 
 return 0; /* Unreachable */
}

